#include "Symbol.hh"
#include "Token.hh"
#include "Rule.hh"
#include "tokenizer.hh"
#include "common.hh"
#include "debug.hh"


extern Synonyms * synonyms;

#define IERR(msg) {\
  cerr << (msg) << "\n";\
  return new Token(func,form,sfunc,sform,lex,bas);\
}

string first(StringVector * v)
{
  if(!v->size()) return "";
  return *(v->begin());
}

/**
 * Parse a line from stdin to construct a Token
 * The line is assumed to be of one of two forms:
 * 1st form (word): {lexical_item} ("{base_form}" {form} @{function} (@#{sentence_level_tag})? )+
 * 2nd form (punctuation): ${punctuation symbol}
 * tag lines that contain $ in the first element are assumed to be punctuation
 * and have form=punc, func=punc inserted.
 * 
 * e.g.: would "would" fin @FAUX @#FS-OBJ
 *       this "this" pron @SUBJ "this" pron @>N
 *       $,
 * 
 * NOTE: returns 0 on end-of-input 
 */
Token * readToken(Token * prev)
{
#ifdef _OLD_STUFF
  StringVector 
    *lex = new StringVector(), 
    *bas = new StringVector(), 
    *form = new StringVector(), 
    *func = new StringVector(), 
    *sent = new StringVector();
#endif
  
  string tmp="";
  vector<Token*>* tokens = new vector<Token*>();
  
  // skip leading newlines.
  char c = 0;
  while(cin.get(c) && c=='\n');
  if(c==0) return NULL; // end-of-input
  cin.putback(c);
  
  //word token or punctuation token
  while(cin.get(c) && c!=' ' && c!='$' && c!='\n') tmp += c;
  if(c=='$') 
    {
      // this is a punctuation token.
      while(cin.get(c) && c!='\n' && c!=' ' && c!=0) if(c!='\\') tmp += c;
      Token * newToken = new Token("punc","punc","","","$"+tmp,"$"+tmp);
      newToken->setPrevious(prev);
      tokens->push_back(newToken);
      
      // skip to end of line.
      if(c!='\n') while (cin.get(c) && c!='\n' && c!=0) ;
    }
  else
    {
      // this is a word token
      string bas, form, func, sfunc, sform;
      
      // lexical item
      string lex = tmp;
      
      if(c==0) IERR("unexpected end-of-input");
      if(c=='\n') IERR("unexpected end-of-line");
      
      cin.get(c);
      
      while(c!=0 && c!='\n')
	{
	  bas="";
	  form="";
	  func="";
	  sfunc="";
	  sform="";
	  
	  // base form
	  tmp="";
	  if(c!='\"') IERR("baseform expected");
	  while(cin.get(c) && c!='\"' && c!='\n') tmp += c;
	  bas = tmp;
	  if(c!='\"') IERR("missing \" at end of baseform");
	  cin.get(c); // ' ' expected
	  if(c==0) IERR("unexpected end-of-input");
	  if(c=='\n') IERR("unexpected end-of-line");
	  if(c!=' ') IERR("error after baseform of token");
	  
	  // form tag, function tag, and sentence-level tag
	  while(cin.get(c) && c!='\"')
	    {
	      tmp="";
	      if(c=='@')
		{
		  if(!cin.get(c)) IERR("unexpected end-of-input");
		  
		  if(c!='#') // function
		    {
		      tmp +=c;
		      while(cin.get(c) && c!=' ' && c!= '\n') tmp += c;
		      func = tmp;
		      if(c==0) IERR("unexpected end-of-input");
		      if(c=='\n') break;
		    }
		  else // sentence-level tag
		    {
		      while(cin.get(c) && c!=' ' && c!= '\n' ) tmp += c;
		      string::size_type pos = tmp.find('-');
		      if(pos == string::npos) 
			{
			  // no '-' found
			  // the tag is considered a sentence level function tag
			  sfunc = tmp; 
			}
		      else 
			{
			  // split into form and func
			  sfunc = string(tmp,pos+1,string::npos);
			  sform = string(tmp,0,pos);
			}
		    }
		  if(c==0) IERR("unexpected end-of-input");
		  if(c=='\n') break;
		}
	      else // form
		{
		  tmp += c;
		  while(cin.get(c) && c!=' ' && c!= '\n') tmp += c;
		  form = tmp;
		  if(c==0) IERR("unexpected end-of-input");
		  if(c=='\n') break;
		}
	    }
	  if(form == "") IERR("No form found in token.");
	  if(func == "") IERR("No function found in token.");
	  
	  Token * newToken = new Token(func,form,sfunc,sform,lex,bas);
	  newToken->setPrevious(prev);
	  tokens->push_back(newToken);
	  
	  // see if any synonyms should be added
	  for(Synonyms::iterator i=synonyms->begin(),z=synonyms->end(); i!=z; ++i)
	    if(newToken->matches((*i)))
	      {
		Token * synToken = (*i)->rewrite(newToken);
		synToken->setPrevious(prev);
		tokens->push_back(synToken);
	      }
	  // BUG: duplicate token synonyms may result of the above.
      	}
    }
#ifdef _OLD_STUFF_
  //lexical item or punctuation symbol
  tmp="";
  while(cin.get(c) && c!=' ' && c!='$') tmp += c;
  if(c=='$') 
    {
      // this is a punctuation token.
      while(cin.get(c) && c!='\n' && c!=' ' && c!=0) if(c!='\\') tmp += c;
      lex->push_back("$" + tmp);
      bas->push_back("$" + tmp);
      form->push_back("punc");
      func->push_back("punc");
      // skip to end of line.
      if(c!='\n') while (cin.get(c) && c!='\n' && c!=0) ;
    }
  else
    {
      lex->push_back(tmp);
      if(c==0) IERR("unexpected end-of-input");
      
      // base form
      tmp="";
      cin.get(c); // '"' expected
      while(cin.get(c) && c!='\"') tmp += c;
      bas->push_back(tmp);
      if(c==0) IERR("unexpected end-of-input");
      cin.get(c); // " " expected
      
      // form tags, function tags, and sentence-level tags
      while(cin.get(c))
	{
	  tmp="";
	  if(c=='@')
	    {
	      if(!cin.get(c)) IERR("unexpected end-of-input");
	      if(c!='#') // function
		{
		  tmp +=c;
		  while(cin.get(c) && c!=' ' && c!= '\n') tmp += c;
		  func->push_back(tmp);
		  if(c==0) IERR("unexpected end-of-input");
		  if(c=='\n') break;
		}
	      else // sentence-level tag
		{
		  while(cin.get(c) && c!=' ' && c!= '\n' ) tmp += c;
		  sent->push_back(tmp);
		  if(c==0) IERR("unexpected end-of-input");
		  if(c=='\n') break;
		}
	    }
	  else // form
	    {
	      tmp += c;
	      while(cin.get(c) && c!=' ' && c!= '\n') tmp += c;
	      form->push_back(tmp);
	      if(c==0) IERR("unexpected end-of-input");
	      if(c=='\n') break;
	    }
	}
    }
  if(form->size() == 0) IERR("No form found in token.");
  if(func->size() == 0) IERR("No function found in token.");
  
  // make X-product : form x func x sent
  vector<Token*>* tokens = new vector<Token*>(); 
  string l = first(lex); // lexical item and
  string b = first(bas); // base form are constant
  // iterate through form, func:
  for(int i=0,z=form->size(); i!=z; ++i)
    for(int j=0,y=func->size(); j!=y; ++j)
      {
	string fu = (*func)[j];
	string fo = (*form)[i];
	string sfu= "";
	string sfo= "";
	if(sent->size() > 0)
	  {
	    // iterate through sent :
	    for(int k=0,x=sent->size(); k!=x; ++k)
	      {
		// split sentence-level tag into form and function
		string s = (*sent)[k];
		string::size_type pos = s.find('-');
		if(pos == string::npos) // no '-' found
		  {
		    sfu = s; // the tag is considered a sentence level function tag
		  }
		else // split into form and func
		  {
		    sfu = string(s,pos+1,string::npos);
		    sfo = string(s,0,pos);
		  }
	      }
	  }
	// else : no sentence-level tags.

	Token * newToken = new Token(fu, fo, sfu, sfo, l, b);
	newToken->setPrevious(prev);
	tokens->push_back( newToken );
	
	// see if any synonyms should be added
	for(Synonyms::iterator i=synonyms->begin(),z=synonyms->end(); i!=z; ++i)
	  if(newToken->matches((*i))) 
	    {
	      Token * synToken = (*i)->rewrite(newToken);
	      synToken->setPrevious(prev);
	      tokens->push_back(synToken);
	    }
	
	// BUG: duplicate token synonyms may result of the above. 
	// BUG: These should be removed.
      }
#endif
  
  if(tokens->size() > 1){
    TokenSet * set = new TokenSet(tokens);
    set->setPrevious(prev);
    return set;
  }
  return tokens->front();
}
